/*
c3hess - Console Chess written in C
Copyright (C) 2023 Kasyanov Nikolay Alexeyevich

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "algebraic_notation.h"


// Converts a tile notation to a position on the board
Position notation_tile_to_position(TileNotation notation) {
    Position pos;
    switch (notation.file) {
        case 'a': {
            pos.column = 0;            
            break;
        }
        case 'b': {
            pos.column = 1;            
            break;
        }
        case 'c': {
            pos.column = 2;            
            break;
        }
        case 'd': {
            pos.column = 3;            
            break;
        }
        case 'e': {
            pos.column = 4;            
            break;
        }
        case 'f': {
            pos.column = 5;            
            break;
        }
        case 'g': {
            pos.column = 6;            
            break;
        }
        case 'h': {
            pos.column = 7;            
            break;
        }
        default: {
            break;
        }
    }
    // pos.row = notation.rank - 1;
    pos.row = (8 - (notation.rank));

    return pos;
}


// Converts notated tile into its string representation
char* notation_tile_to_string(TileNotation tile) {
    char* notation = (char*) malloc(sizeof(char) * 2);
    if (!notation) {
        return NULL;
    }

    notation[0] = tile.file; // col
    notation[1] = (char) tile.rank; // row

    return notation;
}


// Converts notated tile into its string representation on dest
void notation_tile_to_string_cpy(TileNotation tile, char* dest) {
    if (!dest) {
        return;
    }

    char* notation = notation_tile_to_string(tile);
    strcpy(dest, notation);
    free(notation);

    return;
}


// Converts notated tile string into its corresponding structure
TileNotation notation_tile_from_string(const char* tile_string) {
    if (!tile_string) {
        return notation_tile_new('a', 1);
    }

    TileNotation tile;
    tile.file = tile_string[0];
    tile.rank = (unsigned int) (tile_string[1] - '0'); // Convert to proper int

    return tile;
}


// Creates a new notation tile from given file and rank. 
TileNotation notation_tile_new(char file, unsigned int rank) {
    TileNotation tile;
    tile.file = file;
    tile.rank = rank;

    return tile;
}


// Converts position to a notated position
TileNotation notation_tile_from_pos(Position pos) {
    TileNotation tile;    
    switch (pos.row) {
        case 1: {
            tile.file = 'a';            
            break;
        }
        case 2: {
            tile.file = 'b';            
            break;
        }
        case 3: {
            tile.file = 'c';            
            break;
        }
        case 4: {
            tile.file = 'd';            
            break;
        }
        case 5: {
            tile.file = 'e';            
            break;
        }
        case 6: {
            tile.file = 'f';            
            break;
        }
        case 7: {
            tile.file = 'g';            
            break;
        }
        case 8: {
            tile.file = 'h';            
            break;
        }
        default: {
            tile.file = '\0';
            break;
        }
    }
    tile.rank = pos.column;

    return tile;
}



// Returns a piece character. Upper case for white, lower case for black
char notation_piece_to_char(Piece piece) {
    switch (piece.type) {
        case KING: {
            if (piece.color == WHITE) {
                return 'K';
            } else {
                return 'k';
            }
            break;
        }
        case ROOK: {
            if (piece.color == WHITE) {
                return 'R';
            } else {
                return 'r';
            }
            break;
        }
        case BISHOP: {
            if (piece.color == WHITE) {
                return 'B';
            } else {
                return 'b';
            }
            break;
        }
        case KNIGHT: {
            if (piece.color == WHITE) {
                return 'N';
            } else {
                return 'n';
            }
            break;
        }
        case PAWN: {
            if (piece.color == WHITE) {
                return 'P';
            } else {
                return 'p';
            }
            break;
        }
        case QUEEN: {
            if (piece.color == WHITE) {
                return 'Q';
            } else {
                return 'q';
            }
            break;
        }
        default: {
            return '\0';
        }
    };
}


// Returns 1 if character is one of the piece characters, 0 otherwise
int notation_is_piece_char(char character) {
    if (
        character == 'K' || character == 'k' || \
        character == 'R' || character == 'r' || \
        character == 'B' || character == 'b' || \
        character == 'N' || character == 'n' || \
        character == 'P' || character == 'p' || \
        character == 'Q' || character == 'q'
    ) {
        return 1;
    }

    return 0;
}


/*
Returns a piece deduced from character, returns NONE piece if couldn't deduce a piece type.
Big characters are for WHITE pieces, small for BLACK pieces.
*/
Piece notation_piece_from_char(char character) {
    if (character == 'K') return piece_new(WHITE, KING);
    else if (character == 'k') return piece_new(BLACK, KING);
    else if (character == 'R') return piece_new(WHITE, ROOK);
    else if (character == 'r') return piece_new(BLACK, ROOK);
    else if (character == 'B') return piece_new(WHITE, BISHOP);
    else if (character == 'b') return piece_new(BLACK, BISHOP);
    else if (character == 'N') return piece_new(WHITE, KNIGHT);
    else if (character == 'n') return piece_new(BLACK, KNIGHT);
    else if (character == 'P') return piece_new(WHITE, PAWN);
    else if (character == 'p') return piece_new(BLACK, PAWN);
    else if (character == 'Q') return piece_new(WHITE, QUEEN);
    else if (character == 'q') return piece_new(BLACK, QUEEN);

    return piece_new(WHITE, NONE);
}


// Converts notated move into its string representation
char* notation_move_to_string(MoveNotation move) {
    char* move_notation = (char*) malloc(sizeof(char) * 10);
    if (!move_notation) {
        return NULL;
    }

    // if (move.move_type == MOVE) {

    // }

    switch (move.move_type) {
        case MOVE: {
            move_notation[0] = notation_piece_to_char(move.piece);
            notation_tile_to_string_cpy(move.target, &move_notation[1]);
            break;
        }

        case CHECK: {
            move_notation[0] = notation_piece_to_char(move.piece);
            notation_tile_to_string_cpy(move.target, &move_notation[1]);
            move_notation[3] = '+';
            break;
        }

        case CAPTURE: {
            move_notation[0] = notation_piece_to_char(move.piece);
            move_notation[1] = 'x';
            notation_tile_to_string_cpy(move.target, &move_notation[2]);
            break;
        }

        case CHECK & CAPTURE: {
            move_notation[0] = notation_piece_to_char(move.piece);
            move_notation[1] = 'x';
            notation_tile_to_string_cpy(move.target, &move_notation[2]);
            move_notation[4] = '+';
            break;
        }

        case SHORTCASTLE: {
            move_notation = strcpy(move_notation, "0-0");
            break;
        }

        // case SHORTCASTLE&CHECK: {

        // }

        case LONGCASTLE: {
            move_notation = strcpy(move_notation, "0-0-0");
            break;
        }

        default: {
            free(move_notation);
            return NULL;
        }
    }

    return move_notation;
}


/*
Accepts a move string which consists of an origin square, piece name
(upper case for WHITE, lower case for BLACK) and a square to move to 
in algebraic notation.

Returns a notated move with origin, target and piece values set. Move type is to be
determined by the engine.

Returns NULL if technical errors were encountered, move_string does not contain enough information
*/
MoveNotation* notation_move_from_string(const char* move_string) {
    unsigned int len = strlen(move_string);

    if (!move_string || len < 5) {
        return NULL;
    }

    MoveNotation* move = (MoveNotation*) malloc(sizeof(MoveNotation) * 1);
    if (!move) {
        return NULL;
    }


    // Origin tile
    TileNotation origin = notation_tile_from_string(move_string);

    // Piece
    Piece piece = notation_piece_from_char(move_string[2]);
    if (piece.type == NONE) {
        // No piece information given
        return NULL;
    }

    // Target tile
    TileNotation target = notation_tile_from_string(move_string + 3);

    move->origin = origin;
    move->piece = piece;
    move->target = target;
    move->move_type = 0;

    return move;
}


// Creates a new move notation from provided arguments
MoveNotation notation_move_new(Piece piece, TileNotation origin, TileNotation target, int move_type) {
    MoveNotation move_new;
    move_new.piece = piece;
    move_new.origin = origin;
    move_new.target = target;
    move_new.move_type = move_type;

    return move_new;
}


// Converts move notation to a move
Move notation_to_move(MoveNotation notation) {
    Move move = move_new(
        notation_tile_to_position(notation.origin),
        notation_tile_to_position(notation.target)
    );

    return move;
}