/*
c3hess - Console Chess written in C
Copyright (C) 2023 Kasyanov Nikolay Alexeyevich

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/


#include "board.h"
#include "tile.h"
#include "position.h"
#include "util.h"
#include "algebraic_notation.h"
#include <stddef.h>
#include <stdlib.h>



// Returns a newly allocated chess board with pieces placed 
Board* board_new_default() {
    Board* board = NULL;

    board = (Board*) malloc(sizeof(Board) * 1);
    if (!board) {
        return NULL;
    }

    board->rows = 8;
    board->columns = 8;
    board->tiles = (Tile*) malloc(sizeof(Tile) * (board->rows * board->columns));
    if (!board->tiles) {
        free(board);
        return NULL;
    }

    // Initialize the board with empty tiles
    for (int r = 0; r < board->rows; r++) {
        for (int c = 0; c < board->columns; c++) {
            board->tiles[r * board->columns + c] = tile_new_empty(position_new(r, c));
        }
    }

    // Put pieces into their initial positions

    // White
    board_place_piece(board, position_new(0, 0), piece_new(WHITE, ROOK));
    board_place_piece(board, position_new(0, 1), piece_new(WHITE, KNIGHT));
    board_place_piece(board, position_new(0, 2), piece_new(WHITE, BISHOP));
    board_place_piece(board, position_new(0, 3), piece_new(WHITE, KING));
    board_place_piece(board, position_new(0, 4), piece_new(WHITE, QUEEN));
    board_place_piece(board, position_new(0, 5), piece_new(WHITE, BISHOP));
    board_place_piece(board, position_new(0, 6), piece_new(WHITE, KNIGHT));
    board_place_piece(board, position_new(0, 7), piece_new(WHITE, ROOK));
    board_place_piece(board, position_new(1, 0), piece_new(WHITE, PAWN));
    board_place_piece(board, position_new(1, 1), piece_new(WHITE, PAWN));
    board_place_piece(board, position_new(1, 2), piece_new(WHITE, PAWN));
    board_place_piece(board, position_new(1, 3), piece_new(WHITE, PAWN));
    board_place_piece(board, position_new(1, 4), piece_new(WHITE, PAWN));
    board_place_piece(board, position_new(1, 5), piece_new(WHITE, PAWN));
    board_place_piece(board, position_new(1, 6), piece_new(WHITE, PAWN));
    board_place_piece(board, position_new(1, 7), piece_new(WHITE, PAWN));

    // Black 
    board_place_piece(board, position_new(7, 0), piece_new(BLACK, ROOK));
    board_place_piece(board, position_new(7, 1), piece_new(BLACK, KNIGHT));
    board_place_piece(board, position_new(7, 2), piece_new(BLACK, BISHOP));
    board_place_piece(board, position_new(7, 3), piece_new(BLACK, KING));
    board_place_piece(board, position_new(7, 4), piece_new(BLACK, QUEEN));
    board_place_piece(board, position_new(7, 5), piece_new(BLACK, BISHOP));
    board_place_piece(board, position_new(7, 6), piece_new(BLACK, KNIGHT));
    board_place_piece(board, position_new(7, 7), piece_new(BLACK, ROOK));
    board_place_piece(board, position_new(6, 0), piece_new(BLACK, PAWN));
    board_place_piece(board, position_new(6, 1), piece_new(BLACK, PAWN));
    board_place_piece(board, position_new(6, 2), piece_new(BLACK, PAWN));
    board_place_piece(board, position_new(6, 3), piece_new(BLACK, PAWN));
    board_place_piece(board, position_new(6, 4), piece_new(BLACK, PAWN));
    board_place_piece(board, position_new(6, 5), piece_new(BLACK, PAWN));
    board_place_piece(board, position_new(6, 6), piece_new(BLACK, PAWN));
    board_place_piece(board, position_new(6, 7), piece_new(BLACK, PAWN));


    return board;
}


/*
Returns a newly allocated chess board with custom amount of 
rows and columns, without pieces placed
*/ 
Board* board_new_empty_custom(unsigned int rows, unsigned int columns) {
    Board* board = NULL;
    return board;
}


// Frees up board and sets the pointer to it to NULL
void board_delete(Board** board) {
    if (board == NULL) {
        return;
    } else if (*board == NULL) {
        return;
    }

    free((*board)->tiles);
    free(*board);
    *board = NULL;

    return;
}


// Prints board to some FILE
void board_print(const Board* board, FILE* out) {
    if (!board || !out) {
        return;
    }

    // Files
    fputc(' ', out);
    for (unsigned int i = 0; i < board->columns; i++) {
        if (i > 0) {
            fputc(' ', out);
        }
        fputc(0x61 + i, out);
    }
    fputc('\n', out);


    // Upper border
    fputc(' ', out);
    for (unsigned int i = 0; i < board->columns; i++) {
        if (i + 1 != board->columns) {
            fputc('_', out);
        }
        fputc('_', out);
    }
    fputc('\n', out);


    for (unsigned int r = 0; r < board->rows; r++) {
        // Left border
        fputc('|', out);

        for (unsigned int c = 0; c < board->columns; c++) {
            // Tile
            char tile_character = '\0';

            if ((r + c) % 2 == 0) {
                // tile_character = '▓';
                tile_character = '#';
            } else {
                // tile_character = '░';
                tile_character = ' ';
            }

            Tile* tile = board_get_tile(board, position_new(r, c));
            if (tile->piece.type != NONE) {
                tile_character = notation_piece_to_char(tile->piece);
            }

            if (c > 0) {
                fputc(' ', out);
            }
            fputc(tile_character, out);
        }

        // Right border
        fputc('|', out);
        

        // Ranks
        fprintf(out, " %d", board->rows - r);

        fputc('\n', out);
    }

    // Lower border
    fputc(' ', out);
    for (unsigned int i = 0; i < board->columns; i++) {
        if (i + 1 != board->columns) {
            fputc('-', out);
        }
        fputc('-', out);
    }
    fputc('\n', out);
    

    return;
}


// Checks if given tile position is in bounds on this board. Returns 1 if it is, 0 - otherwise
int board_is_tile_in_bounds(const Board* board, Position pos) {
    if (pos.row >= board->rows || pos.column >= board->columns) {
        return 0;
    }

    return 1;
}



// Retrieves a tile on provided position. If available - returns a pointer to it, if not - returns NULL
Tile* board_get_tile_raw(const Board* board, Position pos) {
    if (!board) {
        return NULL;
    }

    if (!board_is_tile_in_bounds(board, pos)) {
        return NULL;
    }

    return &board->tiles[(pos.row * board->columns) + pos.column];
}


/*
Retrieves a tile on provided position as if (0, 0) is the bottom left tile. If available 
- returns a pointer to it, if not - returns NULL.
*/ 
Tile* board_get_tile(const Board* board, Position pos) {
    /*
    Invert position so, for example (0, 0) would be the leftmost 
    tile on the bottom (as it should be)
    */
    Position inverted_pos = position_new(
        _sub_abs(pos.row + 1, board->rows), 
        _sub_abs(pos.column + 1, board->columns)
    );

    return board_get_tile_raw(board, inverted_pos);
}


// Returns all tiles with pieces on them on the board
Tile** board_get_occupied_tiles(const Board* board) {
    if (!board) {
        return NULL;
    }

    unsigned int i = 0;
    Tile** occupied_tiles = (Tile**) malloc(sizeof(Tile) * board->rows * board->columns);
    if (!occupied_tiles) {
        return NULL;
    }

    // Loop through the whole board
    Tile* tile = NULL;
    for (unsigned int r = 0; r < board->rows; r++) {
        for (unsigned int c = 0; c < board->columns; c++) {
            // Always in bounds
            if (board_is_tile_occupied(board, position_new(r, c))) {
                // It is occupied!
                occupied_tiles[i] = tile;
                i++;
            }
        }
    }

    // Realloc
    occupied_tiles = realloc(occupied_tiles, sizeof(Tile) * i);

    return occupied_tiles;
}


// Places a piece on a tile with provided position. Returns 1 if the piece has been placed, 0 if position is out of bounds
int board_place_piece(Board* board, Position pos, Piece piece) {
    if (!board) {
        return 0;
    }

    /*
    Invert position so, for example (0, 0) would be the leftmost 
    tile on the bottom (as it should be)
    */
    Position inverted_pos = position_new(
        _sub_abs(pos.row + 1, board->rows), 
        _sub_abs(pos.column + 1, board->columns)
    );

    Tile* tile = board_get_tile(board, inverted_pos);
    if (!tile) {
        return 0;
    }

    // Place the piece
    tile->piece = piece;

    return 1;
}


// Checks if a tile of given position is occupied or not. Returns 1 if it is occupied, 0 - otherwise
int board_is_tile_occupied(const Board* board, Position pos) {
    if (!board) {
        return 0;
    }

    Tile* tile = board_get_tile(board, pos);
    if (!tile) {
        return 0;
    }

    if (tile->piece.type == NONE) {
        return 0;
    }

    return 1;
}


/*
Returns all possible legal positions in diagonal, in straight lines, in both as well as knight-type for specified tile.

Returns NULL in case of an error
*/
Position* board_get_positions(const Board* board, const Tile* piece_tile, MoveMode mode, unsigned int* pos_count) {
    if (!board || !piece_tile || !pos_count) {
        return NULL;
    }

    unsigned int i = 0;
    Position* legal_moves = (Position*) malloc(sizeof(Position) * board->rows * board->columns);
    if (!legal_moves) {
        return NULL;
    }

    switch (mode) {
        case MODEKNIGHT: {
            for (unsigned int j = 0; j < 8; j++) {
                int vx = KNIGHTDELTAS[j].x;
                int vy = KNIGHTDELTAS[j].y;

                int xtmp = piece_tile->position.row + vx;
                int ytmp = piece_tile->position.column + vy;
                if (board_is_tile_in_bounds(board, position_new(xtmp, ytmp))) {
                    Tile* tile = board_get_tile(board, position_new(xtmp, ytmp));
                    if (tile->piece.type == NONE) {
                        legal_moves[i] = tile->position;
                        i++;
                    } else if (tile->piece.color != piece_tile->piece.color) {
                        legal_moves[i] = tile->position;
                        i++;
                    } else {
                        break;
                    }
                }
            }
            break;
        }

        case MODESTRAIGHTS: {
            for (unsigned int j = 0; j < 4; j++) {
                int vx = CARDINALS[j].x;
                int vy = CARDINALS[j].y;

                int xtmp = piece_tile->position.row + vx;
                int ytmp = piece_tile->position.column + vy;
                while (board_is_tile_in_bounds(board, position_new(xtmp, ytmp))) {
                    Tile* tile = board_get_tile(board, position_new(xtmp, ytmp));
                    if (tile->piece.type == NONE) {
                        legal_moves[i] = tile->position;
                        i++;
                    } else if (tile->piece.color != piece_tile->piece.color) {
                        legal_moves[i] = tile->position;
                        i++;
                        break; // Don't proceed
                    } else {
                        break;
                    }

                    xtmp = xtmp + vx;
                    ytmp = ytmp + vy;
                }
            }
            break;
        }

        case MODEDIAGONALS: {
            for (unsigned int j = 0; j < 4; j++) {
                int vx = DIAGONALS[j].x;
                int vy = DIAGONALS[j].y;

                int xtmp = piece_tile->position.row + vx;
                int ytmp = piece_tile->position.column + vy;
                while (board_is_tile_in_bounds(board, position_new(xtmp, ytmp))) {
                    Tile* tile = board_get_tile(board, position_new(xtmp, ytmp));
                    if (tile->piece.type == NONE) {
                        legal_moves[i] = tile->position;
                        i++;
                    } else if (tile->piece.color != piece_tile->piece.color) {
                        legal_moves[i] = tile->position;
                        i++;
                        break; // Don't proceed
                    } else {
                        break;
                    }

                    xtmp = xtmp + vx;
                    ytmp = ytmp + vy;
                }
            }
            break;
        }

        case MODESTRAIGHTSDIAGONALS: {
            for (unsigned int j = 0; j < 4; j++) {
                int vx = CARDINALS[j].x;
                int vy = CARDINALS[j].y;

                int xtmp = piece_tile->position.row + vx;
                int ytmp = piece_tile->position.column + vy;
                while (board_is_tile_in_bounds(board, position_new(xtmp, ytmp))) {
                    Tile* tile = board_get_tile(board, position_new(xtmp, ytmp));
                    if (tile->piece.type == NONE) {
                        legal_moves[i] = tile->position;
                        i++;
                    } else if (tile->piece.color != piece_tile->piece.color) {
                        legal_moves[i] = tile->position;
                        i++;
                        break; // Don't proceed
                    } else {
                        break;
                    }

                    xtmp = xtmp + vx;
                    ytmp = ytmp + vy;
                }
            }

            for (unsigned int j = 0; j < 4; j++) {
                int vx = DIAGONALS[j].x;
                int vy = DIAGONALS[j].y;

                int xtmp = piece_tile->position.row + vx;
                int ytmp = piece_tile->position.column + vy;
                while (board_is_tile_in_bounds(board, position_new(xtmp, ytmp))) {
                    Tile* tile = board_get_tile(board, position_new(xtmp, ytmp));
                    if (tile->piece.type == NONE) {
                        legal_moves[i] = tile->position;
                        i++;
                    } else if (tile->piece.color != piece_tile->piece.color) {
                        legal_moves[i] = tile->position;
                        i++;
                        break; // Don't proceed
                    } else {
                        break;
                    }

                    xtmp = xtmp + vx;
                    ytmp = ytmp + vy;
                }
            }
            break;
        }

        default: {
            return NULL;
        }
    }

    legal_moves = realloc(legal_moves, sizeof(Position) * i);
    *pos_count = i;

    return legal_moves;
}



// Returns all possible legal knight positions.
Position* board_get_knight_moves(const Board* board, const Tile* knight_tile, unsigned int* pos_count) {
    if (!board || !knight_tile || !pos_count) {
        return NULL;
    }

    return board_get_positions(board, knight_tile, MODEKNIGHT, pos_count);
}


// Returns all possible legal rook moves.
Position* board_get_rook_moves(const Board* board, const Tile* rook_tile, unsigned int* pos_count) {
    if (!board || !rook_tile || !pos_count) {
        return NULL;
    }

    return board_get_positions(board, rook_tile, MODESTRAIGHTS, pos_count);
}


// Returns all possible legal bishop positions.
Position* board_get_bishop_moves(const Board* board, const Tile* bishop_tile, unsigned int* pos_count) {
    if (!board || !bishop_tile || !pos_count) {
        return NULL;
    }

    return board_get_positions(board, bishop_tile, MODEDIAGONALS, pos_count);
}


// Returns all possible legal queen positions.
Position* board_get_queen_moves(const Board* board, const Tile* queen_tile, unsigned int* pos_count) {
    if (!board || !queen_tile || !pos_count) {
        return NULL;
    }

    return board_get_positions(board, queen_tile, MODESTRAIGHTSDIAGONALS, pos_count);
}


// Returns all possible legal king positions.
Position* board_get_king_moves(const Board* board, const Tile* king_tile, unsigned int* pos_count) {
    if (!board || !king_tile || !pos_count) {
        return NULL;
    }

    Position* queen_positions = board_get_queen_moves(board, king_tile, pos_count);
    Position* king_positions = malloc(sizeof(Position) * (*pos_count));
    if (!king_positions) {
        free(queen_positions);
        return NULL;
    }

    unsigned int king_pos_count = 0;
    for (unsigned int i = 0; i < *pos_count; i++) {
        if (position_distance(king_tile->position, queen_positions[i]) > 1) {
            continue;
        }

        king_positions[king_pos_count] = queen_positions[i];
        king_pos_count++;
    }

    free(queen_positions);
    *pos_count = king_pos_count;
    king_positions = realloc(king_positions, sizeof(Position) * king_pos_count);

    return NULL;
}


// Returns all possible legal pawn positions.
Position* board_get_pawn_moves(const Board* board, const Tile* pawn_tile, unsigned int* pos_count) {
    if (!board || !pawn_tile || !pos_count) {
        return NULL;
    }

    
    Position* positions = malloc(sizeof(Position) * board->rows * board->columns);
    if (!positions) {
        return NULL;
    }


    // 1 and 2 tile advancements and takes
    Position pos = pawn_tile->position;
    for (int i = 1; i <= 2; i++) {
        pos = pawn_tile->position;

        // Advances
        if (pawn_tile->piece.color == WHITE) {
            pos.row = pos.row + i;
        } else {
            pos.row = pos.row - i;
        }


        if (board_is_tile_in_bounds(board, pos)) {
            Tile* tile = board_get_tile(board, pos);
            if (tile->piece.type == NONE) {
                // Add this position
                positions[*pos_count] = pos;
                (*pos_count)++;
            }
        }


        // Takes
        if (i % 2 == 0) {
            pos.column = pos.column + 1;
        } else {
            pos.column = pos.column - 1;
        }

        if (board_is_tile_in_bounds(board, pos)) {
            Tile* tile = board_get_tile(board, pos);
            if (tile->piece.type != NONE && tile->piece.color != pawn_tile->piece.color) {
                // Add this position
                positions[*pos_count] = pos;
                (*pos_count)++;
            }
        }
    }


    positions = realloc(positions, sizeof(Position) * (*pos_count));

    return positions;
}


// Returns all possible legal positions for a piece standing on piece_tile. Returns NULL in case of an error
Position* board_get_positions_for(const Board* board, const Tile* piece_tile, unsigned int* pos_count) {
    if (!board || !piece_tile || !pos_count) {
        return NULL;
    }

    // Check if there is a piece or not
    if (piece_tile->piece.type == NONE) {
        return NULL;
    }

    switch (piece_tile->piece.type) {
        case PAWN: {
            return board_get_pawn_moves(board, piece_tile, pos_count);
        }

        case KNIGHT: {
            return board_get_knight_moves(board, piece_tile, pos_count);
        }

        case ROOK: {
            return board_get_rook_moves(board, piece_tile, pos_count);
        }

        case BISHOP: {
            return board_get_bishop_moves(board, piece_tile, pos_count);
        }
        
        case QUEEN: {
            return board_get_queen_moves(board, piece_tile, pos_count);
        }

        case KING: {
            return board_get_king_moves(board, piece_tile, pos_count);
        }
        
        default: {
            return NULL;
        }
    }

    return NULL;
}



/*
Checks if a move is legal or not (and reason) on this board.

Returns MoveLegality
*/
MoveLegality board_is_move_legal(const Board* board, Move move, Color color_plays) {
    // Check if the tiles are not out of bounds
    if (!board_is_tile_in_bounds(board, move.origin) || 
        !board_is_tile_in_bounds(board, move.target)
    ) {
        
        return ILLOUTOFBOUNDS;
    }

    // Check if there is a piece at all
    if (board_get_tile(board, move.origin)->piece.type == NONE) {
        return ILLNOPIECE;
    }

    Tile* origin_tile = board_get_tile(board, move.origin);
    Tile* target_tile = board_get_tile(board, move.target);

    // Check if the piece is of the player's color
    if (origin_tile->piece.color != color_plays) {
        return ILLNOTYOURPIECE;
    }

    // See if the target tile is of the allied color
    if (tile_is_occupied(*target_tile)) {
        // The tile is occupied
        if (target_tile->piece.color == color_plays) {
            // The pieces are allied, can't go there!
            return ILLSAMECOLOR;
        }
    }

    
    // Check based on the piece type
    Piece origin_piece = origin_tile->piece;    
    switch (origin_piece.type) {
        case KING: {
            if (position_distance(move.origin, move.target) > 1) {
                // Too far away
                return ILLINVALIDPIECEMOVEMENT;
            }

            return LEGAL;
        }

        case QUEEN: {
            unsigned int pos_count = 0;
            Position* legal_positions = board_get_queen_moves(
                board, 
                board_get_tile(board, move.origin),
                &pos_count
            );
            
            for (unsigned int i = 0; i < pos_count; i++) {
                if (legal_positions[i].row == move.target.row &&
                    legal_positions[i].column == move.target.column
                ) 
                {
                    // This position is allowed to be moved to
                    free(legal_positions);
                    return LEGAL;
                }
            }

            // Couldn't find such a position in legal placements
            free(legal_positions);
            return ILLINVALIDPIECEMOVEMENT;
        }

        case ROOK: {
            unsigned int pos_count = 0;
            Position* legal_positions = board_get_rook_moves(
                board, 
                board_get_tile(board, move.origin),
                &pos_count
            );
            
            for (unsigned int i = 0; i < pos_count; i++) {
                if (legal_positions[i].row == move.target.row &&
                    legal_positions[i].column == move.target.column
                ) 
                {
                    // This position is allowed to be moved to
                    free(legal_positions);
                    return LEGAL;
                }
            }

            // Couldn't find such a position in legal placements
            free(legal_positions);
            return ILLINVALIDPIECEMOVEMENT;
        }

        case KNIGHT: {
            unsigned int pos_count = 0;
            Position* legal_positions = board_get_knight_moves(
                board, 
                board_get_tile(board, move.origin),
                &pos_count
            );
            
            for (unsigned int i = 0; i < pos_count; i++) {
                if (legal_positions[i].row == move.target.row &&
                    legal_positions[i].column == move.target.column
                ) 
                {
                    // This position is allowed to be moved to
                    free(legal_positions);
                    return LEGAL;
                }
            }

            // Couldn't find such a position in legal placements
            free(legal_positions);
            return ILLINVALIDPIECEMOVEMENT;
        }

        case BISHOP: {
            unsigned int pos_count = 0;
            Position* legal_positions = board_get_bishop_moves(
                board, 
                board_get_tile(board, move.origin),
                &pos_count
            );
            
            for (unsigned int i = 0; i < pos_count; i++) {
                if (legal_positions[i].row == move.target.row &&
                    legal_positions[i].column == move.target.column
                ) 
                {
                    // This position is allowed to be moved to
                    free(legal_positions);
                    return LEGAL;
                }
            }

            // Couldn't find such a position in legal placements
            free(legal_positions);
            return ILLINVALIDPIECEMOVEMENT;
        }

        case PAWN: {
            unsigned int pos_count = 0;
            Position* legal_positions = board_get_pawn_moves(
                board, 
                board_get_tile(board, move.origin),
                &pos_count
            );
            
            for (unsigned int i = 0; i < pos_count; i++) {
                if (legal_positions[i].row == move.target.row &&
                    legal_positions[i].column == move.target.column
                ) 
                {
                    // This position is allowed to be moved to
                    free(legal_positions);
                    return LEGAL;
                }
            }

            // Couldn't find such a position in legal placements
            free(legal_positions);
            return ILLINVALIDPIECEMOVEMENT;
        }

        default: {
            // Non reachable
            return LEGAL;
        }
    }
}


/*
Moves piece from one place to other, does not 
check if move is legal or not.

Returns 1 if move has been successfully done, 0 - otherwise
*/ 
int board_make_move(Board* board, Move move) {
    if (!board) {
        return 0;
    }

    Tile* tile_origin = board_get_tile(board, move.origin);
    if (!tile_origin) {
        return 0;
    }

    Tile* tile_target = board_get_tile(board, move.target);
    if (!tile_target) {
        return 0;
    }

    // Move the piece from one tile to another
    tile_target->piece = tile_origin->piece;
    tile_origin->piece.type = NONE;

    return 1;
}