/*
c3hess - Console Chess written in C
Copyright (C) 2023 Kasyanov Nikolay Alexeyevich

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/


#ifndef board_h
#define board_h

#include "tile.h"
#include "move.h"
#include "position.h"
#include "piece.h"
#include <stdio.h>


typedef struct Vec2 {
    int x;
    int y;
} Vec2;


static const Vec2 CARDINALS[4] = {
    {x: 1, y: 0}, {x: 0, y: 1},
    {x: -1, y: 0}, {x: 0, y: -1}
};


static const Vec2 DIAGONALS[4] = {
    {x: 1, y: 1}, {x: -1, y: 1},
    {x: 1, y: -1}, {x: -1, y: -1}
};


static const Vec2 KNIGHTDELTAS[8] = {
    {x: 1, y: 2}, {x: 2, y: 1},
    {x: -1, y: 2}, {x: -2, y: 1},
    {x: 1, y: -2}, {x: 2, y: -1},
    {x: -1, y: -2}, {x: -2, y: -1}
};


typedef struct Board {
    unsigned int rows;
    unsigned int columns;
    Tile* tiles;
} Board;


// Returns a newly allocated chess board with pieces placed 
Board* board_new_default();


/*
Returns a newly allocated chess board with custom amount of 
rows and columns, without pieces placed
*/ 
Board* board_new_empty_custom(unsigned int rows, unsigned int columns);


// Frees up board and sets the pointer to it to NULL
void board_delete(Board** board);


// Checks if given tile position is in bounds on this board. Returns 1 if it is, 0 - otherwise
int board_is_tile_in_bounds(const Board* board, Position pos);


// Places a piece on a tile with provided position. Returns 1 if the piece has been placed, 0 if position is out of bounds
int board_place_piece(Board* board, Position pos, Piece piece);


// Retrieves a tile on provided position. If available - returns a pointer to it, if not - returns NULL
Tile* board_get_tile_raw(const Board* board, Position pos);


/*
Retrieves a tile on provided position as if (0, 0) is the bottom left tile. If available 
- returns a pointer to it, if not - returns NULL.
*/ 
Tile* board_get_tile(const Board* board, Position pos);


// Returns all tiles with pieces on them on the board
Tile** board_get_occupied_tiles(const Board* board);


// Checks if a tile of given position is occupied or not. Returns 1 if it is occupied, 0 - otherwise
int board_is_tile_occupied(const Board* board, Position pos);



// Prints board to some FILE
void board_print(const Board* board, FILE* out);


typedef enum MoveMode {
    MODESTRAIGHTS,
    MODEDIAGONALS,
    MODESTRAIGHTSDIAGONALS,
    MODEKNIGHT
} MoveMode;


/*
Returns all possible legal positions in diagonal, in straight lines, in both as well as knight-type for specified tile.

Returns NULL in case of an error
*/
Position* board_get_positions(const Board* board, const Tile* piece_tile, MoveMode mode, unsigned int* pos_count);


// Returns all possible legal king positions.
Position* board_get_king_moves(const Board* board, const Tile* king_tile, unsigned int* pos_count);


// Returns all possible legal rook positions.
Position* board_get_rook_moves(const Board* board, const Tile* rook_tile, unsigned int* pos_count);


// Returns all possible legal queen positions.
Position* board_get_queen_moves(const Board* board, const Tile* queen_tile, unsigned int* pos_count);


// Returns all possible legal bishop positions.
Position* board_get_bishop_moves(const Board* board, const Tile* bishop_tile, unsigned int* pos_count);


// Returns all possible legal knight positions.
Position* board_get_knight_moves(const Board* board, const Tile* knight_tile, unsigned int* pos_count);


// Returns all possible legal pawn positions.
Position* board_get_pawn_moves(const Board* board, const Tile* pawn_tile, unsigned int* pos_count);


// Returns all possible legal positions for a piece standing on piece_tile. Returns NULL in case of an error
Position* board_get_positions_for(const Board* board, const Tile* piece_tile, unsigned int* pos_count);


// Move legality status
typedef enum MoveLegality {
    LEGAL,
    ILLSAMECOLOR,
    ILLINVALIDPIECEMOVEMENT,
    ILLOUTOFBOUNDS,
    ILLNOPIECE,
    ILLNOTYOURPIECE
} MoveLegality;


/*
Checks if a move is legal or not (and reason) on this board.

Returns MoveLegality
*/
MoveLegality board_is_move_legal(const Board* board, Move move, Color color_plays);


/*
Moves piece from one place to other, does not 
check if move is legal or not.

Returns 1 if move has been successfully done, 0 - otherwise
*/ 
int board_make_move(Board* board, Move move);

#endif