/*
c3hess - Console Chess written in C
Copyright (C) 2023 Kasyanov Nikolay Alexeyevich

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/


#include "chess.h"
#include "algebraic_notation.h"
#include <stdlib.h>


Chess* chess_new(Options options) {
    if (!options.input || !options.output) {
        return NULL;
    }

    Chess* game = (Chess*) malloc(sizeof(Chess) * 1);
    if (!game) {
        return NULL;
    }

    game->board = board_new_default();
    game->moves_made = 0;
    game->turn = WHITE;
    game->state = READY;
    game->options = options;
    if (options.play_as_white == 1) {
        game->plays_as = WHITE;
    } else {
        game->plays_as = BLACK;
    }

    return game;
}


void chess_delete(Chess** chess) {
    if (!chess) {
        return;
    }

    if (!*chess) {
        return;
    }


    board_delete(&(*chess)->board);
    free(*chess);
    *chess = NULL;

    return;
}


void chess_display(const Chess* chess) {
    char* color_string = color_to_string(chess->turn);
    fprintf(
        chess->options.output,
        "[%s to move; %d moves made]\n\n",
        color_string,
        chess->moves_made    
    );
    free(color_string);

    board_print(chess->board, chess->options.output);
    fprintf(chess->options.output, "\n");
    
    return;
}


Move* chess_take_move_input(Chess* chess) {
    if (!chess) {
        return NULL;
    }

    // Ask for input
    fprintf(chess->options.output, "Move: ");
    char* input = (char*) malloc(sizeof(char) * 64);
    input = fgets(input, 64, chess->options.input);

    // Try to parse input
    MoveNotation* move_notation = notation_move_from_string(input);
    if (!move_notation) {
        // Failed!
        return NULL;
    }

    // Convert it to a useable move
    Move* move = (Move*) malloc(sizeof(Move) * 1);
    if (!move) {
        free(move_notation);
        return NULL;
    }
    *move = notation_to_move(*move_notation);
    free(move_notation);

    return move;
}


void chess_toggle_turn(Chess* chess) {
    if (!chess) {
        return;
    }

    if (chess->turn == WHITE) {
        chess->turn = BLACK;
    } else {
        chess->turn = WHITE;
    }

    return;
}


void chess_next_turn(Chess* chess) {
    if (!chess) {
        return;
    }

    // Keep asking for a legitimate move
    Move* move = NULL;
    MoveLegality result = 0;
    while (1) {
        // Get a move
        move = chess_take_move_input(chess);
        if (!move) {
            printf("Invalid move input! (opt, where o - origin tile, p - piece, t - target tile: a2Pa4 moves white pawn at a2 to a4)\n");
            continue;
        }
        
        printf("(%d %d) --- (%d %d)\n", move->origin.row, move->origin.column, move->target.row, move->target.column);
        
        
        // Check if it's a valid move
        result = board_is_move_legal(chess->board, *move, chess->turn);
        if (result != LEGAL) {
            // TODO: output special message for each illegal move
            
            printf("Illegal move!\n");
            free(move);
            continue;
        }


        // Make a move!
        int move_made = board_make_move(chess->board, *move);
        if (!move_made) {
            printf("Failed to make a move!\n"); // Should not be reacheable
            free(move);
            continue;
        }

        break;
    }
    free(move);

    chess->moves_made++;
    chess_toggle_turn(chess);

    return;
}